from wire_rxtr import RadixTree
import typing

class BaseRouter:
    """
    Here is the routing magic.
    """
    __slots__ = ["routes"]

    def __init__(self, strict: bool = True) -> None:
        self.strict_mode: bool = strict
        self.routes = RadixTree()

    def add_route(self, path: str, handler, method: str, dependencies: list = []) -> bool:
        if not self.strict_mode:
            path = path[:-1] if path.endswith("/") and len(path) > 1 else path
        self.routes.insert(path, handler, method, dependencies)
        return True


class Router(BaseRouter):
    def __init__(self, strict: bool = True):
        super().__init__(strict=strict)

    async def get_handler(self, path: str, method: str) -> typing.Tuple[
        typing.Union[typing.Callable, typing.Awaitable], typing.Optional[dict]]:
        if not self.strict_mode:
            path = path[:-1] if path.endswith("/") and len(path) > 1 else path
        pathFound, handler, params, dependencies = self.routes.get(path, method)
        return handler, params, dependencies



from urllib.parse import parse_qsl
import json
from typing import Optional, ByteString, Union
from multidict import CIMultiDict
from wire.typings import CoroutineFunction
from http.cookies import SimpleCookie
from enum import Enum


class WSState(Enum):
    """This enum is used to represent the websocket state"""
    init = "init"
    connecting = "connecting"
    open = "open"
    closing = "closing"
    closed = "closed"


class ConnectionType(Enum):
    """This enum is used to represent the connection type"""
    http: str = "http"
    ws: str = "websocket"


class Request:
    """This class represents a request from the client to the server"""
    def __init__(
        self,
        scope: dict,
        receive: Optional[CoroutineFunction] = None,
        send: Optional[CoroutineFunction] = None,
    ) -> None:
        self._receive = receive
        self._send = send
        self._scope = scope
        self._req_headers: Optional[CIMultiDict] = None
        self._req_cookies: SimpleCookie = SimpleCookie()
        self.http_body = b""
        self.http_has_more_body = True
        self.http_received_body_length = 0

    @property
    def path(self) -> str:
        """Displays the path that the client uses"""
        return self._scope["path"]

    @property
    def method(self) -> str:
        """Represents the method that the client uses when making the request."""
        return self._scope["method"].lower()

    @property
    def client(self) -> str:
        """Specifies the address and port through which the client connects to the server. ex. (addr,port)"""
        return self._scope["client"]

    @property
    def server(self) -> str:
        """Retrieves the address of the server to which the CLient connects."""
        return self._scope["server"]

    @property
    def headers(self) -> CIMultiDict:
        """Retrieves all headers sent by the client during the request"""
        if not self._req_headers:
            self._req_headers = CIMultiDict(
                [(k.decode("ascii"), v.decode("ascii")) for (k, v) in self._scope["headers"]])
        return self._req_headers

    @property
    def cookies_raw(self) -> SimpleCookie:
        """Retrieves all cookies sent by the client during the request"""
        if not self._req_headers:
            self._req_headers = CIMultiDict(
                [(k.decode("ascii"), v.decode("ascii")) for (k, v) in self._scope["headers"]])
            self._req_cookies.load(self._req_headers.get("cookie", {}))
        return self._req_cookies

    @property
    def cookies(self) -> dict:
        """Retrieves all cookies sent by the client during the request as dict"""
        return {key: m.value for key, m in self.cookies_raw.items()}

    @property
    def scope(self) -> dict:
        return self._scope

    @property
    def query(self) -> dict:
        """Retrieves the querystring as dict"""
        return CIMultiDict(parse_qsl(self.scope.get("query_string", b"").decode("utf-8")))

    @property
    def type(self) -> ConnectionType:
        """retrieves the connection type of the request"""
        return ConnectionType.ws if self.scope.get("type") == "websocket" else ConnectionType.http

    async def body_iter(self):
        if not self.type == ConnectionType.http:
            raise Exception("ConnectionType is not http")
        if self.http_received_body_length > 0 and self.http_has_more_body:
            raise Exception("body iter is already started and is not finished")
        if self.http_received_body_length > 0 and not self.http_has_more_body:
            yield self.http_body
        req_body_length = (
            int(self.headers.get("content-length", "0"))
            if not self.headers.get("transfer-encoding") == "chunked"
            else None
        )
        while self.http_has_more_body:
            if req_body_length and self.http_received_body_length > req_body_length:
                raise Exception("body is longer than excepted")
            message = await self._receive()
            message_type = message.get("type")
            await self.handle(message)
            if message_type != "http.request":
                continue
            chunk: bytes = message.get("body", b"")
            if not isinstance(chunk, bytes):
                raise RuntimeError("Chunk is not bytes")
            self.http_body += chunk
            self.http_has_more_body = message.get("more_body", False)
            self.http_received_body_length += len(chunk)
            yield bytes(chunk)

    async def body(self) -> bytes:
        """Retrieves the request body"""
        data: bytes = b"".join([chunks async for chunks in self.body_iter()])
        try:
            return json.loads(data)
        except json.decoder.JSONDecodeError:
            return data

    async def handle(self, message):
        if message.get("type") == "http.disconnect":
            raise Exception("Disconnected")


from . import Pluggable
from wire.enums import Events
import typing
import inspect
import asyncio

class Eventable(Pluggable):
    def __init__(self) -> None:
        self.events: dict = {}

    def __call__(self, event: str) -> None:
        if isinstance(event, Events):
            event = event.value
        for function in self.events.get(event, []):
            if inspect.iscoroutinefunction(function):
                asyncio.run(function)
            else: 
                function()

    def add(self, event: typing.Union[Events, str] , function: typing.Callable) -> None:
        if isinstance(event, Events):
            event = event.value
        if function in self.events.get(event, []):
            return
        if event not in self.events.keys():
            self.events[event] = [function]
        else: 
            self.events[event].append(function)




from wire.enums import HttpStatusCode
import typing
import json
from fuchs import FuchsTemplate
from http.cookies import SimpleCookie, BaseCookie


class Response:
    """
    This is the base answer class. All other answer classes are based on it.
    """
    media_type = None
    charset = "utf-8"

    def __init__(self, content: typing.Any, status: typing.Union[int, HttpStatusCode] = 200, headers: dict = None,
                 media_type: str = None) -> None:
        self.status = status if isinstance(status, int) else status.value
        self.content: typing.Any = content
        if media_type:
            self.media_type = media_type
        self.body = self.render(content)
        self.set_headers(headers)

    def render(self, content: typing.Any):
        if content is None:
            return b""
        if isinstance(content, bytes):
            return content
        return content.encode(self.charset)

    def set_headers(self, headers: typing.Mapping[str, str] = None) -> None:
        if headers is None:
            raw_headers: typing.List[typing.Tuple[bytes, bytes]] = []
        else:
            raw_headers = [(k.lower().encode(self.charset), v.encode(
                self.charset)) for k, v in headers.items()]
        body: bytes = getattr(self, "body", b"")
        if body:
            raw_headers.append(
                (b"content-length", str(len(body)).encode(self.charset)))
        ctype: str = self.media_type
        if ctype is not None:
            if ctype.startswith("text/"):
                ctype += ";" + "charset=" + self.charset + ";"
            raw_headers.append((b"content-type", ctype.encode(self.charset)))
            # raw_headers.append((b"set-cookie", bytes(morsel.OutputString(), "utf-8")))
        self.headers = raw_headers

    def set_cookie(self, key: str, value: str = "", max_age: int = None,
                   expires: int = None, path: str = "/", domain: str = None,
                   secure: bool = False,
                   httponly: bool = False,
                   samesite: str = "lax") -> None:
        cookie: BaseCookie = SimpleCookie()
        cookie[key] = value
        if max_age:
            cookie[key]["max-age"] = max_age
        if expires:
            cookie[key]["expires"] = expires
        if path:
            cookie[key]["path"] = path
        if domain:
            cookie[key]["domain"] = domain
        if secure:
            cookie[key]["secure"] = True
        if httponly:
            cookie[key]["httponly"] = True
        if samesite is not None:
            assert samesite.lower() in [
                "strict",
                "lax",
                "none",
            ], "samesite must be 'strict', 'lax' or 'none'"
            cookie[key]["samesite"] = samesite
        cookie_val = cookie.output(header="").strip()
        self.headers.append((b"set-cookie", cookie_val.encode("latin-1")))

    async def __call__(self, scope, receive, send) -> None:
        await send({"type": "http.response.start", "status": self.status, "headers": self.headers})
        await send({"type": "http.response.body", "body": self.body})


class HTMLResponse(Response):
    media_type = "text/html"


class JsonResponse(Response):
    media_type = "application/json"

    def render(self, content: typing.Any):
        return json.dumps(content, indent=None).encode(self.charset)


class PlainTextResponse(Response):
    media_type = "text/plain"




from wire.response import Response, HTMLResponse, JsonResponse, PlainTextResponse
from wire.request import Request
from wire.static import StaticFiles
from . import Pluggable
import inspect
import typing
import pydantic


class Alternator(Pluggable):

    async def __call__(self, req: Request, func: typing.Callable[[], typing.Any], params: dict, deps: list) -> Response:
        # if function is none, return
        if func is None:
            return PlainTextResponse("Not Found", 404)
        # arguments
        arguments: dict = dict()
        # get args of function
        args = [(name, typ.annotation) for name, typ in list(inspect.signature(func).parameters.items())]

        # replace arg with request object, if arg[0][1] is instance of Request
        for arg in args:
            if arg[1] == Request:
                arguments[arg[0]] = req
            elif issubclass(arg[1], pydantic.BaseModel):
                cls = arg[1]
                arguments[arg[0]] = cls(**(await req.body()))

        # update arguments with params
        arguments.update(params)


        if inspect.iscoroutinefunction(func):
            response: typing.Union[Response, typing.Any] = await func(**arguments)
            if not isinstance(response, Response):
                if type(response) == str:
                    response = PlainTextResponse(response, 200)
                elif type(response) == dict:
                    response = JsonResponse(response, 200)
            return response
        
        # static files
        elif isinstance(func, StaticFiles):
            resp: Response = await func(arguments["filename"])
            return resp

        # sync function
        else:
            response: typing.Union[Response, typing.Any] = func(**arguments)
            if not isinstance(response, Response):
                if type(response) == str:
                    response = PlainTextResponse(response, 200)
                elif type(response) == dict:
                    response = JsonResponse(response, 200)
            return response
        

        
        


from wire.groups.group import Group
from wire.plugs import Pluggable
from wire.plugs.alternator import Alternator
from wire.router import Router
from wire.request import Request
from wire.static import StaticFiles
from wire.listener import EventListener
from wire.enums import Events
import inspect
import typing
import traceback

class Wire:
    def __init__(
        self,
        title: str = "Wire APP",
        description: str = "",
        strict: bool = False,
    ) -> None:
        self.title: str = title
        self.description: str = description
        self.router: Router = Router(strict=strict)
        self.__event_listener: EventListener = EventListener()
        self.alternator = Alternator()

    async def __call__(self, scope: dict, receive, send) -> None:
        # lifespan handling
        try:
            if scope["type"] == "lifespan":
                while True:
                    message = await receive()
                    if message['type'] == 'lifespan.startup':
                        self.__event_listener(Events.ASGI_STARTUP)
                        await send({'type': 'lifespan.startup.complete'})
                    elif message['type'] == 'lifespan.shutdown':
                        self.__event_listener(Events.ASGI_STARTUP)
                        await send({'type': 'lifespan.shutdown.complete'})
                        return
            else:
                req = Request(scope, receive, send)
                func, params, deps = await self.router.get_handler(req.path, req.method)
                resp = await self.alternator(req, func, params, deps)
                await resp(scope, receive, send)
        except:
            traceback.print_exc()
            await send({'type': 'http.response.start', 'status': 500})
            await send({'type': 'http.response.body', 'body': b"Internal Server Error"})
                
                    

    def get(self, path: str, dependencies: list = []):
        def wrapper(handler: typing.Callable) -> typing.Callable:
            self.router.add_route(path, handler, "get", dependencies)
            return handler

        return wrapper

    def post(self, path: str, dependencies: list = []):
        def wrapper(handler: typing.Callable) -> typing.Callable:
            self.router.add_route(path, handler, "post", dependencies)
            return handler

        return wrapper

    def put(self, path: str, dependencies: list = []):
        def wrapper(handler: typing.Callable) -> typing.Callable:
            self.router.add_route(path, handler, "put", dependencies)
            return handler

        return wrapper

    def delete(self, path: str, dependencies: list = []):
        def wrapper(handler: typing.Callable) -> typing.Callable:
            self.router.add_route(path, handler, "delete", dependencies)
            return handler

        return wrapper

    def mount(self, plugin: Pluggable, prefix: str = None) -> bool:
        self.__event_listener(Events.PLUGIN_MOUNT)
        if not prefix.startswith("/"):
            raise Exception(f"Invalid prefix for router {plugin}")
        if isinstance(plugin, StaticFiles):
            self.router.add_route(prefix + "/*filename", plugin, "get")
            self.__event_listener(Events.STATIC_MOUNT)
            return True
        if isinstance(plugin, Group):
            for route in plugin.routes:
                path, handler, method, dependencies = route
                if not self.router.strict_mode and path.startswith("/"):
                    path = path[:-1]

                self.router.add_route(prefix + path, handler, method, dependencies)
            self.__event_listener(Events.GROUP_MOUNT)
            return True
            
    
    def event(self, event: str) -> None:
        def wrapper(handler: typing.Union[typing.Callable, typing.Awaitable]) -> None:
            self.__event_listener.add(event, handler)
            return handler

        return wrapper



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wire', 'wire.groups', 'wire.plugs']

package_data = \
{'': ['*']}

install_requires = \
['multidict>=5.1.0,<6.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'wire-fuchs>=0.1.0,<0.2.0',
 'wire-rxtr>=0.2,<0.3']

entry_points = \
{'console_scripts': ['run = ']}

setup_kwargs = {
    'name': 'wire-web',
    'version': '0.3.2',
    'description': 'Wire the web',
    'long_description': '<h1 align="center">Wire</h1>\n<p align="center"> Easy, fast, stable </p>\n<img src="https://cdn.discordapp.com/attachments/857979752991031296/928221250520760330/wire1.png" align="right" style="margin-top: -50px;"/>\n<br>\n<br>\nWire is designed to provide the user with the greatest possible comfort when creating Rest APIs or entire websites.\nEverything is simple and, above all, intuitively designed. No focus on superfluous configurations. Everything works, simply.\n\n🔑 Key features\n\n- intuitive, due to the clear design\n- simple, due to the fast learning curve\n- practical, through the great editor support\n- minimalistic, no superfluous functions\n\n#### What is Wire and what is not\n\nWire is not a HighSpeed framework. Wire is probably not ready for production. Wire is a spare time project of mine. Wire is self-contained. It doesn\'t need anything, except for an ASGI server. So it\'s like Starlette.\nI would appreciate if you use Wire, try it and give me your feedback.\n\n#### Participate in Wire\n\nYou are welcome to collaborate on Wire. However, you should maintain the codestyle, and also follow PEP 8 (the Python style guide).\n\n#### Wire disadvantages\n\nWire is still deep in development, which is why some features are still missing. \n\n- Websockets\n\n#### Examples\n\nHere is the most basic example of wire\n\n```py\nfrom wire import Wire, Request\n\napp = Wire()\n\n@app.get("/home")\nasync def home(req: Request):\n\treturn "Welcome home"\n```\n\nYou want to build a RestAPI? No problem\n\n```py\nfrom wire import Wire, Request\n\n\napp = Wire()\ntemplates = FoxTemplates("templates")\n\n@app.get("/api")\ndef api(req: Request):\n\treturn {"name": "Leo", "age": 16}\n```\n\nYou want to send HTML files? Wire got your back\n\n```py\nfrom wire import Wire, Request\nfrom wire.responses import HTMLResponse\n\n\napp = Wire()\n\n@app.get("/html")\nasync def home(req: Request):\n\twith open("home.html", "r") as f:\n\t\tdata = f.read()\n\treturn HTMLResponse(data)\n```\n\nYou want to use some templates ? You want to load templates? No problem with [Fuchs](https://github.com/cheetahbyte/fuchs)\n\n```py\nfrom wire import Wire, Request\nfrom wire.templating import FoxTemplates\n\napp = Wire()\ntemplates = FoxTemplates("templates")\n\n@app.get("/home")\nasync def home(req: Request):\n\treturn templates.render("home.html", name="Leo")\n```\n\n**Changes incoming**\n\nJoin our [discord](https://discord.gg/EtqGfBVuZS) !\n',
    'author': 'Leo B.',
    'author_email': 'bernerdoodle@outlook.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://wire.bernerdev.de/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
